@@ -28,11 +28,13 @@ import (
 
 	"github.com/gophercloud/gophercloud"
 	"github.com/gophercloud/gophercloud/openstack"
+	"github.com/gophercloud/gophercloud/openstack/blockstorage/v3/volumes"
 	"github.com/gophercloud/gophercloud/openstack/common/extensions"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/attachinterfaces"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/bootfromvolume"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/floatingips"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/extensions/keypairs"
+	"github.com/gophercloud/gophercloud/openstack/compute/v2/flavors"
 	"github.com/gophercloud/gophercloud/openstack/compute/v2/servers"
 	"github.com/gophercloud/gophercloud/openstack/identity/v3/tokens"
 	"github.com/gophercloud/gophercloud/openstack/imageservice/v2/images"
@@ -44,6 +46,7 @@ import (
 	"github.com/gophercloud/gophercloud/openstack/networking/v2/subnets"
 	"github.com/gophercloud/gophercloud/pagination"
 	"github.com/gophercloud/utils/openstack/clientconfig"
+	configclient "github.com/openshift/client-go/config/clientset/versioned/typed/config/v1"
 	machinev1 "github.com/openshift/cluster-api/pkg/apis/machine/v1beta1"
 	"github.com/openshift/cluster-api/pkg/util"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -59,6 +62,15 @@ const (
 
 	TimeoutPortDelete       = 3 * time.Minute
 	RetryIntervalPortDelete = 5 * time.Second
+
+	// MachineRegionLabelName as annotation name for a machine region
+	MachineRegionLabelName = "machine.openshift.io/region"
+
+	// MachineAZLabelName as annotation name for a machine AZ
+	MachineAZLabelName = "machine.openshift.io/zone"
+
+	// MachineInstanceTypeLabelName as annotation name for a machine instance type
+	MachineInstanceTypeLabelName = "machine.openshift.io/instance-type"
 )
 
 type InstanceService struct {
@@ -67,6 +79,9 @@ type InstanceService struct {
 	identityClient *gophercloud.ServiceClient
 	networkClient  *gophercloud.ServiceClient
 	imagesClient   *gophercloud.ServiceClient
+	volumeClient   *gophercloud.ServiceClient
+
+	regionName string
 }
 
 type Instance struct {
@@ -91,6 +106,19 @@ type InstanceListOpts struct {
 	Name string `q:"name"`
 }
 
+type serverMetadata struct {
+	// AZ contains name of the server's availability zone
+	AZ string `json:"OS-EXT-AZ:availability_zone"`
+
+	// Flavor refers to a JSON object, which itself indicates the hardware
+	// configuration of the deployed server.
+	Flavor map[string]interface{} `json:"flavor"`
+
+	// Status contains the current operational status of the server,
+	// such as IN_PROGRESS or ACTIVE.
+	Status string `json:"status"`
+}
+
 func GetCloudFromSecret(kubeClient kubernetes.Interface, namespace string, secretName string, cloudName string) (clientconfig.Cloud, error) {
 	emptyCloud := clientconfig.Cloud{}
 
@@ -104,7 +132,7 @@ func GetCloudFromSecret(kubeClient kubernetes.Interface, namespace string, secre
 
 	secret, err := kubeClient.CoreV1().Secrets(namespace).Get(secretName, metav1.GetOptions{})
 	if err != nil {
-		return emptyCloud, err
+		return emptyCloud, fmt.Errorf("Failed to get secrets from kubernetes api: %v", err)
 	}
 
 	content, ok := secret.Data[CloudsSecretKey]
@@ -125,7 +153,7 @@ func GetCloudFromSecret(kubeClient kubernetes.Interface, namespace string, secre
 func NewInstanceServiceFromMachine(kubeClient kubernetes.Interface, machine *machinev1.Machine) (*InstanceService, error) {
 	machineSpec, err := openstackconfigv1.MachineSpecFromProviderSpec(machine.Spec.ProviderSpec)
 	if err != nil {
-		return nil, err
+		return nil, fmt.Errorf("Failed to get Machine Spec from Provider Spec (clients/machineservice.go 138): %v", err)
 	}
 	cloud := clientconfig.Cloud{}
 	if machineSpec.CloudsSecret != nil && machineSpec.CloudsSecret.Name != "" {
@@ -135,7 +163,7 @@ func NewInstanceServiceFromMachine(kubeClient kubernetes.Interface, machine *mac
 		}
 		cloud, err = GetCloudFromSecret(kubeClient, namespace, machineSpec.CloudsSecret.Name, machineSpec.CloudName)
 		if err != nil {
-			return nil, err
+			return nil, fmt.Errorf("Failed to get cloud from secret (clients/machienservice.go 150): %v", err)
 		}
 	}
 	return NewInstanceServiceFromCloud(cloud)
@@ -198,12 +226,21 @@ func NewInstanceServiceFromCloud(cloud clientconfig.Cloud) (*InstanceService, er
 		return nil, fmt.Errorf("Create ImageClient err: %v", err)
 	}
 
+	volumeClient, err := openstack.NewBlockStorageV3(provider, gophercloud.EndpointOpts{
+		Region: clientOpts.RegionName,
+	})
+	if err != nil {
+		return nil, fmt.Errorf("Create VolumeClient err: %v", err)
+	}
+
 	return &InstanceService{
 		provider:       provider,
 		identityClient: identityClient,
 		computeClient:  serverClient,
 		networkClient:  networkingClient,
 		imagesClient:   imagesClient,
+		volumeClient:   volumeClient,
+		regionName:     clientOpts.RegionName,
 	}, nil
 }
 
@@ -299,11 +336,12 @@ func getSubnetsByFilter(is *InstanceService, opts *subnets.ListOpts) ([]subnets.
 	return snets, nil
 }
 
-func CreatePort(is *InstanceService, name string, net ServerNetwork, securityGroups *[]string) (ports.Port, error) {
+func CreatePort(is *InstanceService, name string, net ServerNetwork, securityGroups *[]string, allowedAddressPairs *[]ports.AddressPair) (ports.Port, error) {
 	portCreateOpts := ports.CreateOpts{
-		Name:           name,
-		NetworkID:      net.networkID,
-		SecurityGroups: securityGroups,
+		Name:                name,
+		NetworkID:           net.networkID,
+		SecurityGroups:      securityGroups,
+		AllowedAddressPairs: *allowedAddressPairs,
 	}
 	if net.subnetID != "" {
 		portCreateOpts.FixedIPs = []ports.IP{{SubnetID: net.subnetID}}
@@ -384,8 +422,7 @@ func getImageID(is *InstanceService, imageName string) (string, error) {
 }
 
 // InstanceCreate creates a compute instance
-func (is *InstanceService) InstanceCreate(clusterName string, name string, clusterSpec *openstackconfigv1.OpenstackClusterProviderSpec, config *openstackconfigv1.OpenstackProviderSpec, cmd string, keyName string) (instance *Instance, err error) {
-	var createOpts servers.CreateOptsBuilder
+func (is *InstanceService) InstanceCreate(clusterName string, name string, clusterSpec *openstackconfigv1.OpenstackClusterProviderSpec, config *openstackconfigv1.OpenstackProviderSpec, cmd string, keyName string, configClient configclient.ConfigV1Interface) (instance *Instance, err error) {
 	if config == nil {
 		return nil, fmt.Errorf("create Options need be specified to create instace")
 	}
@@ -451,6 +488,30 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 			}
 		}
 	}
+	if len(nets) == 0 {
+		return nil, fmt.Errorf("No network was found or provided. Please check your machine configuration and try again")
+	}
+
+	clusterInfra, err := configClient.Infrastructures().Get("cluster", metav1.GetOptions{})
+	if err != nil {
+		return nil, fmt.Errorf("Failed to retrieve cluster Infrastructure object: %v", err)
+	}
+
+	allowedAddressPairs := []ports.AddressPair{}
+	if clusterInfra != nil && clusterInfra.Status.PlatformStatus != nil && clusterInfra.Status.PlatformStatus.OpenStack != nil {
+		clusterVips := []string{
+			clusterInfra.Status.PlatformStatus.OpenStack.APIServerInternalIP,
+			clusterInfra.Status.PlatformStatus.OpenStack.NodeDNSIP,
+			clusterInfra.Status.PlatformStatus.OpenStack.IngressIP,
+		}
+
+		for _, vip := range clusterVips {
+			if vip != "" {
+				allowedAddressPairs = append(allowedAddressPairs, ports.AddressPair{IPAddress: vip})
+			}
+		}
+	}
+
 	userData := base64.StdEncoding.EncodeToString([]byte(cmd))
 	var ports_list []servers.Network
 	for _, net := range nets {
@@ -471,7 +532,7 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 		var port ports.Port
 		if len(portList) == 0 {
 			// create server port
-			port, err = CreatePort(is, name, net, &securityGroups)
+			port, err = CreatePort(is, name, net, &securityGroups, &allowedAddressPairs)
 			if err != nil {
 				return nil, fmt.Errorf("Failed to create port err: %v", err)
 			}
@@ -538,7 +599,7 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 		return nil, fmt.Errorf("Create new server err: %v", err)
 	}
 
-	serverCreateOpts := servers.CreateOpts{
+	var serverCreateOpts servers.CreateOptsBuilder = servers.CreateOpts{
 		Name:             name,
 		ImageRef:         imageID,
 		FlavorName:       config.Flavor,
@@ -556,21 +617,75 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 	if config.RootVolume != nil && config.RootVolume.Size != 0 {
 		var blocks []bootfromvolume.BlockDevice
 
+		volumeID := config.RootVolume.SourceUUID
+
+		// change serverCreateOpts to exclude imageRef from them
+		serverCreateOpts = servers.CreateOpts{
+			Name:             name,
+			FlavorName:       config.Flavor,
+			AvailabilityZone: config.AvailabilityZone,
+			Networks:         ports_list,
+			UserData:         []byte(userData),
+			SecurityGroups:   securityGroups,
+			ServiceClient:    is.computeClient,
+			Tags:             serverTags,
+			Metadata:         config.ServerMetadata,
+			ConfigDrive:      config.ConfigDrive,
+		}
+
+		if bootfromvolume.SourceType(config.RootVolume.SourceType) == bootfromvolume.SourceImage {
+			// if source type is "image" then we have to create a volume from the image first
+			klog.Infof("Creating a bootable volume from image %v.", config.RootVolume.SourceUUID)
+
+			imageID, err := getImageID(is, config.RootVolume.SourceUUID)
+			if err != nil {
+				return nil, fmt.Errorf("Create new server err: %v", err)
+			}
+
+			// Create a volume first
+			volumeCreateOpts := volumes.CreateOpts{
+				Size:       config.RootVolume.Size,
+				VolumeType: config.RootVolume.VolumeType,
+				ImageID:    imageID,
+				// The same name as the instance
+				Name: name,
+			}
+
+			volume, err := volumes.Create(is.volumeClient, volumeCreateOpts).Extract()
+			if err != nil {
+				return nil, fmt.Errorf("Create bootable volume err: %v", err)
+			}
+
+			volumeID = volume.ID
+
+			err = volumes.WaitForStatus(is.volumeClient, volumeID, "available", 300)
+			if err != nil {
+				klog.Infof("Bootable volume %v creation failed. Removing...", volumeID)
+				err = volumes.Delete(is.volumeClient, volumeID, volumes.DeleteOpts{}).ExtractErr()
+				if err != nil {
+					return nil, fmt.Errorf("Bootable volume deletion err: %v", err)
+				}
+
+				return nil, fmt.Errorf("Bootable volume %v is not available err: %v", volumeID, err)
+			}
+
+			klog.Infof("Bootable volume %v was created successfully.", volumeID)
+		}
+
 		block := bootfromvolume.BlockDevice{
-			SourceType:          bootfromvolume.SourceType(config.RootVolume.SourceType),
+			SourceType:          bootfromvolume.SourceVolume,
 			BootIndex:           0,
-			UUID:                config.RootVolume.SourceUUID,
+			UUID:                volumeID,
 			DeleteOnTermination: true,
 			DestinationType:     bootfromvolume.DestinationVolume,
-			VolumeSize:          config.RootVolume.Size,
-			DeviceType:          config.RootVolume.DeviceType,
 		}
 		blocks = append(blocks, block)
 
-		createOpts = bootfromvolume.CreateOptsExt{
-			CreateOptsBuilder: createOpts,
+		serverCreateOpts = bootfromvolume.CreateOptsExt{
+			CreateOptsBuilder: serverCreateOpts,
 			BlockDevice:       blocks,
 		}
+
 	}
 
 	server, err := servers.Create(is.computeClient, keypairs.CreateOptsExt{
@@ -580,6 +695,7 @@ func (is *InstanceService) InstanceCreate(clusterName string, name string, clust
 	if err != nil {
 		return nil, fmt.Errorf("Create new server err: %v", err)
 	}
+
 	is.computeClient.Microversion = ""
 	return serverToInstance(server), nil
 }
@@ -706,6 +822,38 @@ func (is *InstanceService) GetInstance(resourceId string) (instance *Instance, e
 	return serverToInstance(server), err
 }
 
+// SetMachineLabels set labels describing the machine
+func (is *InstanceService) SetMachineLabels(machine *machinev1.Machine, instanceID string) error {
+	if machine.Labels[MachineRegionLabelName] != "" && machine.Labels[MachineAZLabelName] != "" && machine.Labels[MachineInstanceTypeLabelName] != "" {
+		return nil
+	}
+
+	var sm serverMetadata
+	err := servers.Get(is.computeClient, instanceID).ExtractInto(&sm)
+	if err != nil {
+		return err
+	}
+
+	if machine.Labels == nil {
+		machine.Labels = make(map[string]string)
+	}
+
+	// Set the region
+	machine.Labels[MachineRegionLabelName] = is.regionName
+
+	// Set the availability zone
+	machine.Labels[MachineAZLabelName] = sm.AZ
+
+	// Set the flavor name
+	flavor, err := flavors.Get(is.computeClient, sm.Flavor["id"].(string)).Extract()
+	if err != nil {
+		return err
+	}
+	machine.Labels[MachineInstanceTypeLabelName] = flavor.Name
+
+	return nil
+}
+
 func serverToInstance(server *servers.Server) *Instance {
 	return &Instance{*server}
 }
